structure Curses = 
    struct 
	datatype color = black | red | green | brown | darkyellow | blue | purple | cyan | lightgray 
	  | lightgrey | darkgray | darkgrey | lightred | lightgreen | lightbrown | yellow | lightblue 
	  | lightpurple | lightcyan | white;
	val esc = chr 27;
			   
	(* raw output *)
	fun rawWrite x = 
	    Posix.IO.writeVec (Posix.FileSys.stdout, {buf=x, i=0, sz=SOME(Word8Vector.length x)});

	(* setup for raw output *)
	fun charListtoWord8Vec_helper nil = nil
	  | charListtoWord8Vec_helper (a::d) = Word8.fromInt(ord(a))::charListtoWord8Vec_helper d;

	fun charListtoWord8Vec x = Word8Vector.fromList (charListtoWord8Vec_helper x);

	(* convenience functions *)
	fun writeCharList x = rawWrite (charListtoWord8Vec x);
	fun writeString x = writeCharList (String.explode x);
	fun putch x = writeCharList [x];

	(* Color control *)

	fun setForegroundColor col = 
	    (writeCharList ([esc, #"["]@
			    (case col of
				 black => [#"0", #";", #"3", #"0"]
			       | red => [#"0", #";", #"3", #"1"]
			       | green => [#"0", #";", #"3", #"2"]
			       | brown => [#"0", #";", #"3", #"3"]
			       | darkyellow => [#"0", #";", #"3", #"3"]
			       | blue => [#"0", #";", #"3", #"4"]
			       | purple => [#"0", #";", #"3", #"5"]
			       | cyan => [#"0", #";", #"3", #"6"]
			       | lightgray => [#"0", #";", #"3", #"7"]
			       | lightgrey => [#"0", #";", #"3", #"7"]
			       | darkgray => [#"1", #";", #"3", #"0"]
			       | darkgrey => [#"1", #";", #"3", #"0"]
			       | lightred => [#"1", #";", #"3", #"1"]
			       | lightgreen => [#"1", #";", #"3", #"2"]
			       | lightbrown => [#"1", #";", #"3", #"3"]
			       | yellow => [#"1", #";", #"3", #"3"]
			       | lightblue => [#"1", #";", #"3", #"4"]
			       | lightpurple => [#"1", #";", #"3", #"5"]
			       | lightcyan => [#"1", #";", #"3", #"6"]
			       | white => [#"1", #";", #"3", #"7"])@[#"m"]); ())

	fun setBackgroundColor col = 
	    (writeCharList ([esc, #"["]@
			    (case col of
				 black => [#"4", #"0"]
			       | red => [#"4", #"1"]
			       | green => [#"4", #"2"]
			       | brown => [#"4", #"3"]
			       | darkyellow => [#"4", #"3"]
			       | blue => [#"4", #"4"]
			       | purple => [#"4", #"5"]
			       | cyan => [#"4", #"6"]
			       | lightgray => [#"4", #"7"]
			       | lightgrey => [#"4", #"7"]
			       | darkgray => [#"4", #"0"]
			       | darkgrey => [#"4", #"0"]
			       | lightred => [#"4", #"1"]
			       | lightgreen => [#"4", #"2"]
			       | lightbrown => [#"4", #"3"]
			       | yellow => [#"4", #"3"]
			       | lightblue => [#"4", #"4"]
			       | lightpurple => [#"4", #"5"]
			       | lightcyan => [#"4", #"6"]
			       | white => [#"4", #"7"])@[#"m"]); ())



	fun bold ()      = (writeCharList [esc, #"[", #"1", #"m"]; ());
	fun plain ()     = (writeCharList [esc, #"[", #"0", #"m"]; ());

	val setForegroundColour = setForegroundColor;
	(* display stuff *)
	fun gotoxy x y = (writeCharList ([esc, #"["]
					 @String.explode(Int.toString(y+1))
					 @[#";"]
					 @String.explode(Int.toString(x+1))
					 @[#"H"]); ())
	fun clear () = (writeCharList [esc, #"[", #"2", #"J"];
		       gotoxy 0 0; ())

    (* Input stuff *)
	fun lowerflag flags flagout =
	    Posix.TTY.L.fromWord (Word32.andb (Posix.TTY.L.toWord flags, 
					       Word32.notb (Posix.TTY.L.toWord flagout)));
	fun raiseflag flags flagout =
	    Posix.TTY.L.fromWord (Word32.orb (Posix.TTY.L.toWord flags, 
					      (Posix.TTY.L.toWord flagout)));
(*
	fun rawmodeflaghelper _ nil = nil
	  | rawmodeflaghelper lf (a::d) = 
	    if Posix.TTY.L.anySet (a, lf) then
		a::(rawmodeflaghelper lf d)
	    else (rawmodeflaghelper lf d);
*)
	fun rawmode () = 
	    let 
		val attr = Posix.TTY.getattr Posix.FileSys.stdin;
		val fields = Posix.TTY.fieldsOf attr;
		val lflag = Posix.TTY.getlflag attr;
	    in
		Posix.TTY.setattr (Posix.FileSys.stdin, 
				   Posix.TTY.TC.sanow,
				   Posix.TTY.termios {cc=(#cc(fields)),
						      cflag=(#cflag(fields)),
						      iflag=(#iflag(fields)),
						      ispeed=(#ispeed(fields)),
						      oflag=(#oflag(fields)),
						      ospeed=(#ospeed(fields)),
						      lflag=(lowerflag lflag Posix.TTY.L.icanon)
						      (*
						       lflag=(Posix.TTY.L.flags (rawmodeflaghelper 
						       lflag 
						       [Posix.TTY.L.echo,
						       Posix.TTY.L.echoe,
						       Posix.TTY.L.echok,
						       Posix.TTY.L.echonl,
						       Posix.TTY.L.iexten,
						       Posix.TTY.L.isig,
						       Posix.TTY.L.noflsh,
						       Posix.TTY.L.tostop]))
						       *)
						      })
	    end;

	fun cookedmode () = 
	    let 
		val attr = Posix.TTY.getattr Posix.FileSys.stdin;
		val fields = Posix.TTY.fieldsOf attr;
		val lflag = Posix.TTY.getlflag attr;
	    in
		Posix.TTY.setattr (Posix.FileSys.stdin, 
				   Posix.TTY.TC.sanow,
				   Posix.TTY.termios {cc=(#cc(fields)),
						      cflag=(#cflag(fields)),
						      iflag=(#iflag(fields)),
						      ispeed=(#ispeed(fields)),
						      oflag=(#oflag(fields)),
						      ospeed=(#ospeed(fields)),
						      lflag=(raiseflag lflag Posix.TTY.L.icanon)
						      (*
						       lflag=(Posix.TTY.L.flags ((rawmodeflaghelper 
						       lflag 
						       [Posix.TTY.L.echo,
						       Posix.TTY.L.echoe,
						       Posix.TTY.L.echok,
						       Posix.TTY.L.echonl,
						       Posix.TTY.L.iexten,
						       Posix.TTY.L.isig,
						       Posix.TTY.L.noflsh,
						       Posix.TTY.L.tostop])
						       @[Posix.TTY.L.icanon]))
						       *)
						      })
	    end;

	fun echo_off () = 
	    let 
		val attr = Posix.TTY.getattr Posix.FileSys.stdin;
		val fields = Posix.TTY.fieldsOf attr;
		val lflag = Posix.TTY.getlflag attr;
	    in
		Posix.TTY.setattr (Posix.FileSys.stdin, 
				   Posix.TTY.TC.sanow,
				   Posix.TTY.termios {cc=(#cc(fields)),
						      cflag=(#cflag(fields)),
						      iflag=(#iflag(fields)),
						      ispeed=(#ispeed(fields)),
						      oflag=(#oflag(fields)),
						      ospeed=(#ospeed(fields)),
						      lflag=(lowerflag lflag Posix.TTY.L.echo)})
	    end;

	fun echo_on () = 
	    let 
		val attr = Posix.TTY.getattr Posix.FileSys.stdin;
		val fields = Posix.TTY.fieldsOf attr;
		val lflag = Posix.TTY.getlflag attr;
	    in
		Posix.TTY.setattr (Posix.FileSys.stdin, 
				   Posix.TTY.TC.sanow,
				   Posix.TTY.termios {cc=(#cc(fields)),
						      cflag=(#cflag(fields)),
						      iflag=(#iflag(fields)),
						      ispeed=(#ispeed(fields)),
						      oflag=(#oflag(fields)),
						      ospeed=(#ospeed(fields)),
						      lflag=(raiseflag lflag Posix.TTY.L.echo)})
	    end;


	fun invis_cursor () =
	    writeCharList [esc, #"[", #"?", #"2", #"5", #"l"]

	fun vis_cursor () =
	    writeCharList [esc, #"[", #"?", #"2", #"5", #"h"]

	(* raw output *)
	fun rawRead x = 
	    Posix.IO.readVec (Posix.FileSys.stdout, x);

	(* setup for raw output *)
	fun Word8VectoCharList_helper v n = 
	    if Word8Vector.length(v) = n then nil
	    else chr(Word8.toInt(Word8Vector.sub (v, n)))::(Word8VectoCharList_helper v (n+1));
	fun Word8VectoCharList v = Word8VectoCharList_helper v 0;	    

	(* Convenience functions *)
	fun readCharList x = Word8VectoCharList (rawRead x);
	fun readString x = String.implode (readCharList x);
	fun getch () = List.hd (Word8VectoCharList (rawRead 1));

(*
 use "../mines/curses.sml";
	    Posix.TTY.L.anySet (Posix.TTY.L.icanon,  Posix.TTY.getlflag (Posix.TTY.getattr Posix.FileSys.stdin));
*)
    end;